@@ -1,10 +1,11 @@
 package Lemonldap::NG::Handler::Main::Reload;
 
-our $VERSION = '2.0.12';
+our $VERSION = '2.19.0';
 
 package Lemonldap::NG::Handler::Main;
 
 use strict;
+use Digest::SHA qw(sha256_hex);
 use Lemonldap::NG::Common::Conf::Constants;    #inherits
 use Lemonldap::NG::Common::Crypto;
 use Lemonldap::NG::Common::Safelib;            #link protected safe Safe object
@@ -65,7 +66,8 @@ sub checkConf {
         or $class->cfgNum != $conf->{cfgNum}
         or $class->cfgDate != $conf->{cfgDate} )
     {
-        $class->logger->debug("Get configuration $conf->{cfgNum}");
+        $class->logger->debug(
+            "Get configuration $conf->{cfgNum} aged $conf->{cfgDate}");
         unless ( $class->cfgNum( $conf->{cfgNum} )
             && $class->cfgDate( $conf->{cfgDate} ) )
         {
@@ -82,6 +84,9 @@ sub checkConf {
                     $class->logger->error( "Underlying object can't load conf ("
                           . ref($obj)
                           . "->$sub)" );
+                    if ( $obj->can('error') ) {
+                        $Lemonldap::NG::Common::Conf::msg = $obj->error;
+                    }
                     return 0;
                 }
             }
@@ -204,18 +209,20 @@ sub jailInit {
 sub defaultValuesInit {
     my ( $class, $conf ) = @_;
 
-    $class->tsv->{$_} = $conf->{$_} foreach ( qw(
-        cookieExpiration        cookieName         customFunctions httpOnly
+    $class->tsv->{$_} = $conf->{$_}
+      foreach ( qw(
+        cookieExpiration        cookieName         customFunctions
         securedCookie           timeout            timeoutActivity
         timeoutActivityInterval useRedirectOnError useRedirectOnForbidden
-        useSafeJail             whatToTrace        handlerInternalCache
+        useSafeJail  httpOnly   whatToTrace        handlerInternalCache
         handlerServiceTokenTTL  customToTrace      lwpOpts lwpSslOpts
-        authChoiceAuthBasic     authChoiceParam    hiddenAttributes
-        upgradeSession
+        authChoiceAuthBasic     authChoiceParam    upgradeSession
+        hashedSessionStore
         )
-    );
+      );
 
     $class->tsv->{cipher} = Lemonldap::NG::Common::Crypto->new( $conf->{key} );
+    $class->tsv->{keyH}   = sha256_hex( $conf->{key} );
 
     foreach my $opt (qw(https port maintenance)) {
 
@@ -248,6 +255,8 @@ sub defaultValuesInit {
               $conf->{vhostOptions}->{$vhost}->{vhostServiceTokenTTL};
             $class->tsv->{accessToTrace}->{$vhost} =
               $conf->{vhostOptions}->{$vhost}->{vhostAccessToTrace};
+            $class->tsv->{devOpsRulesUrl}->{$vhost} =
+              $conf->{vhostOptions}->{$vhost}->{vhostDevOpsRulesUrl};
         }
     }
     return 1;
@@ -518,14 +527,16 @@ sub conditionSub {
             $url
             ? (
                 sub {
-                    $_[1]->{_logout} = $url;
+                    my ( $req, $session ) = @_;
+                    $session->{_logout} = $url;
                     return 0;
                 },
                 0
               )
             : (
                 sub {
-                    $_[1]->{_logout} = $class->tsv->{portal}->();
+                    my ( $req, $session ) = @_;
+                    $session->{_logout} = $class->tsv->{portal}->($req);
                     return 0;
                 },
                 0
@@ -533,55 +544,14 @@ sub conditionSub {
         );
     }
 
-    # Since filter exists only with Apache>=2, logout_app and logout_app_sso
-    # targets are available only for it.
-    # This error can also appear with Manager configured as CGI script
-    if ( $cond =~ /^logout_app/i
-        and not $class->isa('Lemonldap::NG::Handler::ApacheMP2::Main') )
-    {
-        $class->logger->info(
-            "Rules logout_app and logout_app_sso require Apache>=2");
-        return ( sub { 1 }, 0 );
-    }
-
     # logout_app
     if ( $cond =~ /^logout_app(?:\s+(.*))?$/i ) {
-        my $u = $1 || $class->tsv->{portal}->();
-        eval 'use Apache2::Filter' unless ( $INC{"Apache2/Filter.pm"} );
-        return (
-            sub {
-                $_[0]->{env}->{'psgi.r'}->add_output_filter(
-                    sub {
-                        return $class->redirectFilter( $u, @_ );
-                    }
-                );
-                1;
-            },
-            0
-        );
+        return $class->logout_app($1);
     }
+
+    # logout_app_sso
     elsif ( $cond =~ /^logout_app_sso(?:\s+(.*))?$/i ) {
-        my $u = $1 || $class->tsv->{portal}->();
-        eval 'use Apache2::Filter' unless ( $INC{"Apache2/Filter.pm"} );
-        return (
-            sub {
-                my ($req) = @_;
-                $class->localUnlog( $req, @_ );
-                $req->{env}->{'psgi.r'}->add_output_filter(
-                    sub {
-                        my $r = $_[0]->r;
-                        return $class->redirectFilter(
-                            &{ $class->tsv->{portal} }() . "?url="
-                              . $class->encodeUrl( $req, $u )
-                              . "&logout=1",
-                            @_
-                        );
-                    }
-                );
-                1;
-            },
-            0
-        );
+        return $class->logout_app_sso($1);
     }
 
     my $mayskip = 0;
@@ -599,6 +569,15 @@ sub conditionSub {
     return ( $sub, $mayskip );
 }
 
+## These methods should be overriden in subclasses
+sub logout_app {
+    return ( sub { 1 }, 0 );
+}
+
+sub logout_app_sso {
+    return ( sub { 1 }, 0 );
+}
+
 ## @method arrayref aliasInit
 # @param options vhostOptions configuration item
 # @return arrayref of vhost and aliases
@@ -639,7 +618,7 @@ sub oauth2Init {
 
 sub substitute {
     my ( $class, $expr ) = @_;
-    $expr ||= '';
+    $expr //= '';
 
     # substitute special vars, just for retro-compatibility
     $expr =~ s/\$date\b/&date/sg;
@@ -656,7 +635,14 @@ sub substitute {
     $expr =~ s/\binGroup\(([^)]*)\)/listMatch(\$s->{'hGroups'},$1,1)/g;
 
     # handle has2f
-    $expr =~ s/\bhas2f\(([^),]*)\)/has2f(\$s,$1)/g;
+    $expr =~ s/\bhas2f\(([^),]*)\)/has2f_internal(\$s,$1)/g;
+
+    # handle inSubnet
+    $expr =~ s/\binSubnet\(([^)]*)\)/ipInSubnet(\$r->{env}->{REMOTE_ADDR},$1)/g;
+
+    # handle inSubnet
+    $expr =~
+      s/\binDomain\(([^)]*)\)/inDomain_internal(\$r->{env}->{HTTP_HOST},$1)/g;
 
     return $expr;
 }
